/*
 * SMPDataFlowAnalysis.h - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#ifndef SMPDATAFLOWANALYSIS_H
#define SMPDATAFLOWANALYSIS_H 1

// SMPDataFlowAnalysis.h
//
// This header defines the common enumerations and types needed by
//  instructions, basic blocks, and functions.

#include <list>
#include <vector>
#include <map>
#include <set>
#include <string>

#include <cstddef>

#include <pro.h>
#include <ua.hpp>

#include "SMPDBInterface.h"
using namespace std;

class SMPFunction;
class SMPBasicBlock;
class SMPInstr;
class DefOrUseList;
class SMPPhiFunction;
class STARSBitSet;

typedef op_t STARSOpndType;

// IDA_SDK_VERSION is defined in pro.h starting in version 5.2 or 5.7 of the SDK.
//  If it is not defined, just assume version 5.1 to make sure we don't try to
//  use features that were added in version 6.0.
#ifndef IDA_SDK_VERSION
#define IDA_SDK_VERSION 610
#endif

// Insert a floating no-op instruction at top of each function to hold SSA DEFs
//  of LiveIn names?
#define SMP_USE_SSA_FNOP_MARKER 1

// Should we try to propagate memory location types until we encounter
//  an indirect memory access or other potential alias in the DEF-USE chain?
#define SMP_PROPAGATE_MEM_TYPES 1

// A maximum string length for use when SMP needs to use less space than
//  the IDA Pro MAXSTR, which is 1024 at present.
#define MAXSMPSTR 256
// Instruction disassembly text.
#define MAXSMPDISASMSTR 192
// Global var names
#define MAXSMPGLOBALVARSTR 64
// Local var names
#define MAXSMPVARSTR 32
// String for dead registers in SMPInstr
#define MAXSMPDEADREGSSTR 48

// Value for an SSA subscript number before it is initialized by SSA renaming.
#define SMP_SSA_UNINIT  (-1)

// Map register number to a string for printing or annotations.
extern const char *RegNames[];
const char *MDGetRegName(STARSOpndType RegOp); // Distinguishes subword regs from their parent regs
extern const unsigned char RegSizes[]; // size in bytes of each reg

// Define a machine-dependent constant for the EFLAGS register in the x86 architecture.
//  IDA Pro distinguishes among four different flag bits; we aggregate them for most purposes,
//  but separate them for SCCP (Sparse Conditional Constant Propagation).
#define X86_FLAGS_REG R_efl
#define MD_FLAGS_REG X86_FLAGS_REG

// Define various important registers for machine-dependent use.
#define X86_CARRY_FLAG R_cf
#define MD_CARRY_FLAG X86_CARRY_FLAG

#define X86_ZERO_FLAG R_zf
#define MD_ZERO_FLAG X86_ZERO_FLAG

#define X86_SIGN_FLAG R_sf
#define MD_SIGN_FLAG X86_SIGN_FLAG

#define X86_OVERFLOW_FLAG R_of
#define MD_OVERFLOW_FLAG X86_OVERFLOW_FLAG

#define X86_INSTRUCTION_POINTER_REG R_ip
#define MD_INSTRUCTION_POINTER_REG X86_INSTRUCTION_POINTER_REG

#define X86_STACK_POINTER_REG R_sp
#define MD_STACK_POINTER_REG X86_STACK_POINTER_REG

#define X86_FRAME_POINTER_REG R_bp
#define MD_FRAME_POINTER_REG X86_FRAME_POINTER_REG

#define X86_RETURN_VALUE_REG R_ax
#define MD_RETURN_VALUE_REG X86_RETURN_VALUE_REG

#define X86_LAST_REG_NO R_last
#define MD_LAST_REG_NO X86_LAST_REG_NO

#define X86_CALL_INSTRUCTION NN_call
#define MD_CALL_INSTRUCTION X86_CALL_INSTRUCTION

#define X86_MOVE_INSTRUCTION NN_mov
#define MD_MOVE_INSTRUCTION X86_MOVE_INSTRUCTION

#define X86_DEFAULT_RETURN_ADDRESS_SIZE STARS_ISA_Bytewidth
#define MD_DEFAULT_RETURN_ADDRESS_SIZE X86_DEFAULT_RETURN_ADDRESS_SIZE

#define X86_BINARY_NOP_INSTRUCTION 0x90
#define MD_BINARY_NOP_INSTRUCTION X86_BINARY_NOP_INSTRUCTION

// Part of our solution to the dense encoding of x86-64 is to record some instruction
//  prefix and auxprefix info in each operand, so we can continue to do data flow analysis
//  on operands long after the instructions containing the operands have gone stale.
//  IDA Pro has macros that operate on their global variable 'cmd' which is only current
//  while analyzing an inst, and is too big to pass around with all operands in Phi DEFs,
//  LiveIn and LiveOut sets, SSA numbering and operand comparisons, etc. So, we record
//  the insnpref byte (a.k.a. cmd.rex) into each operand in the specflag4 field, then we
//  record the following two bits of info into unused bits in specflag4.
const int STARS_VEXPR = 0x40;            // IDA Pro vexpr() was true on inst containing current op_t
const int STARS_VSIB  = 0x20;            // IDA Pro is_vsib() was true on inst containing current op_t

// Debug: print one operand from an instruction or DEF or USE list.
void PrintDefUse(ulong feature, int OpNum);
void PrintSIB(STARSOpndType Opnd);
void AnnotPrintSIB(STARSOpndType Opnd, bool HasOffset, FILE *OutFile);
void SPARKAnnotPrintSIB(STARSOpndType Opnd, bool HasOffset, FILE *OutFile, uint16 SegReg, bool UseFP);
void PrintOneOperand(STARSOpndType Opnd, uint32 features, int OpNum); 
void PrintListOperand(STARSOpndType Opnd, int SSANum = SMP_SSA_UNINIT); 
void PrintOperand(STARSOpndType Opnd);

// Annotations: concisely print one operand.
void AnnotPrintOperand(STARSOpndType Opnd, FILE *OutFile);

// Print opcode string.
void PrintOpcode(uint16 opcode, FILE *OutFile);

// MACHINE DEPENDENT: Could operand be an indirect memory access?
bool MDIsIndirectMemoryOpnd(STARSOpndType CurrOp, bool UseFP);

// MACHINE DEPENDENT: Extract base reg from SIB byte of operand.
int MD_STARS_sib_base(const STARSOpndType &x);

// MACHINE DEPENDENT: Extract index reg from SIB byte of operand.
regnum_t MD_STARS_sib_index(const STARSOpndType &x);

// Is CurrOp a memory operand?
bool IsMemOperand(STARSOpndType CurrOp);

// MACHINE DEPENDENT: Is CurrOp the flags register?
bool MDIsFlagsReg(STARSOpndType CurrOp);

// MACHINE DEPENDENT: Is operand a stack memory access?
bool MDIsStackAccessOpnd(STARSOpndType CurrOp, bool UseFP);

// MACHINE DEPENDENT: Is operand a direct stack memory access?
bool MDIsDirectStackAccessOpnd(STARSOpndType CurrOp, bool UseFP);

// MACHINE DEPENDENT: Is operand trackable in data flow analyses (i.e. a stack memory access or a register?)
bool MDIsDataFlowOpnd(STARSOpndType CurrOp, bool UseFP);

// MACHINE DEPENDENT: Is register a stack pointer or frame pointer?
bool MDIsStackPtrReg(int RegNumber, bool UseFP);

// MACHINE DEPENDENT: Is operand a caller-saved register?
bool MDIsCallerSavedReg(STARSOpndType CurrOp);

// MACHINE DEPENDENT: Get base and index registers and displacement/addr from operand.
void MDExtractAddressFields(STARSOpndType MemOp, int &BaseReg, int &IndexReg, ushort &Scale, ea_t &Offset);

// MACHINE DEPENDENT: Is operand type a known type that we want to analyze?
bool MDKnownOperandType(STARSOpndType TempOp);

// MACHINE DEPENEDENT: Convert subword register to its enclosing register.
ushort MDCanonicalizeSubReg(const ushort Reg1);

// If TempOp is a register, call MDCanonicalizeSubReg() on it.
void CanonicalizeOpnd(STARSOpndType &TempOp);

// Is RegOp the stack pointer or frame pointer register
bool MDIsStackOrFramePointerReg(STARSOpndType RegOp, bool UseFP);

// MACHINE DEPENDENT: Ordering function for register enum values, to use in set containers.
bool MDLessReg(const ushort Reg1, const ushort Reg2);

// MACHINE DEPENDENT: Ordering function for register enum values, to use in set containers.
bool MDEqReg(const ushort Reg1, const ushort Reg2);

// MACHINE DEPENDENT: Ordering function for register operands, to use in set containers.
bool MDLessRegOpnd(const STARSOpndType RegOp1, const STARSOpndType RegOp2);

// Is CurrOp a general purpose register? (not flags, instruction pointer, non-integer reg, etc.)
bool MDIsGeneralPurposeReg(STARSOpndType CurrOp);

// Hash a global name and SSA number into an int, for use in SMPFunction.GlobalDefAddrBySSA map
int HashGlobalNameAndSSA(STARSOpndType DefOp, int SSANum);

// Are operands equal?
bool IsEqOp(STARSOpndType Opnd1, STARSOpndType Opnd2);

// Are operands equal, ignoring bitwidth differences for register operands?
bool IsEqOpIgnoreBitwidth(STARSOpndType Opnd1, STARSOpndType Opnd2);

// Comparison class to use in sorted containers of IDA Pro addresses (type ea_t).
class LessAddr {
public:
	bool operator()(const ea_t &Addr1, const ea_t &Addr2) const {
		return (Addr1 < Addr2);
	}
};

// Comparison class for sorted containers of register operands to search for.
//  First item in each pair is a code address where we are about to search, and the
//  second item in each pair is a hash of SSANum and register number obtained from
//  the HashGlobalNameAndSSA() function.
class LessSearchOperand {
public:
	bool operator()(const pair<ea_t, int> &SearchPair1, const pair<ea_t, int> &SearchPair2) const {
		return ((SearchPair1.first < SearchPair2.first)
			|| ((SearchPair1.first == SearchPair2.first) && (SearchPair1.second < SearchPair2.second)));
	}
};

// MACHINE DEPENDENT: comparison class to permit sorting of STARSOpndType operands.
class LessOp {
public:
	bool operator()(const STARSOpndType &Opnd1, const STARSOpndType &Opnd2) const {
		if (Opnd1.type != Opnd2.type)
			return (Opnd1.type < Opnd2.type);
		switch (Opnd1.type) {
			case o_void: return false;
			case o_reg: return MDLessReg(Opnd1.reg, Opnd2.reg);
			case o_mem: return (Opnd1.addr < Opnd2.addr);
			case o_phrase: if (Opnd1.hasSIB && Opnd2.hasSIB) return ((Opnd1.sib < Opnd2.sib) || ((Opnd1.sib == Opnd2.sib) && (Opnd1.specflag4 < Opnd2.specflag4)));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return MDLessReg(Opnd1.phrase, Opnd2.phrase); // no SIB bytes
			case o_displ: if (Opnd1.hasSIB && Opnd2.hasSIB)
							  return ((Opnd1.sib < Opnd2.sib) 
								|| ((Opnd1.sib == Opnd2.sib) 
									&& ((Opnd1.addr < Opnd2.addr) || ((Opnd1.addr == Opnd2.addr) && (Opnd1.specflag4 < Opnd2.specflag4)))));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return ((Opnd1.addr < Opnd2.addr)
								|| ((Opnd1.addr == Opnd2.addr) 
								&& MDLessReg(Opnd1.reg, Opnd2.reg))); // no SIB bytes
			case o_imm: return (Opnd1.value < Opnd2.value);
			case o_far:  // fall through to o_near case
			case o_near: return (Opnd1.addr < Opnd2.addr);
			case o_trreg:  // fall through
			case o_dbreg:  // fall through
			case o_crreg:  // fall through
			case o_fpreg:  // fall through
			case o_mmxreg: // fall through
			case o_xmmreg: return (Opnd1.reg < Opnd2.reg); // no subword regs to deal with

			default: msg("ERROR: Unknown operand type in LessOp.\n"); return false;
		}; // end switch (Opnd1.type)
	} // end operator
}; // end class LessOp

// MACHINE DEPENDENT: comparison class to permit sorting of definitions, each of which is a pair<STARSOpndType, ea_t>.
class LessDefinition {
public:
	bool operator()(const pair<STARSOpndType, ea_t> &Defn1, const pair<STARSOpndType, ea_t> &Defn2) const {
		STARSOpndType Opnd1 = Defn1.first;
		STARSOpndType Opnd2 = Defn2.first;
		if (IsEqOp(Opnd1, Opnd2)) {
			// If definitions are of the same operand, sort by instruction address of the definition.
			return (Defn1.second < Defn2.second);
		}
		if (Opnd1.type != Opnd2.type)
			return (Opnd1.type < Opnd2.type);
		switch (Opnd1.type) {
			case o_void: return false;
			case o_reg: return MDLessRegOpnd(Opnd1, Opnd2);
			case o_mem: return (Opnd1.addr < Opnd2.addr);
			case o_phrase: if (Opnd1.hasSIB && Opnd2.hasSIB) return ((Opnd1.sib < Opnd2.sib) || ((Opnd1.sib == Opnd2.sib) && (Opnd1.specflag4 < Opnd2.specflag4)));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return MDLessReg(Opnd1.phrase, Opnd2.phrase); // no SIB bytes
			case o_displ: if (Opnd1.hasSIB && Opnd2.hasSIB)
							  return ((Opnd1.sib < Opnd2.sib) 
								|| ((Opnd1.sib == Opnd2.sib) 
									&& ((Opnd1.addr < Opnd2.addr) || ((Opnd1.addr == Opnd2.addr) && (Opnd1.specflag4 < Opnd2.specflag4)))));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return ((Opnd1.addr < Opnd2.addr)
								|| ((Opnd1.addr == Opnd2.addr) 
								&& MDLessReg(Opnd1.reg, Opnd2.reg))); // no SIB bytes
			case o_imm: return (Opnd1.value < Opnd2.value);
			case o_far:  // fall through to o_near case
			case o_near: return (Opnd1.addr < Opnd2.addr);
			case o_trreg:  // fall through
			case o_dbreg:  // fall through
			case o_crreg:  // fall through
			case o_fpreg:  // fall through
			case o_mmxreg: // fall through
			case o_xmmreg: return (Opnd1.reg < Opnd2.reg); // no subword regs to deal with

			default: msg("ERROR: Unknown operand type in LessDefinition.\n"); return false;
		}; // end switch (Opnd1.type)
	} // end operator
}; // end class LessDefinition

// MACHINE DEPENDENT: comparison class to permit sorting of stack delta definitions, each of which is a pair<STARSOpndType, pair<ea_t, sval_t> >.
class LessStackDeltaCopy {
public:
	bool operator()(const pair<STARSOpndType, pair<ea_t, sval_t> > &DeltaCopy1, const pair<STARSOpndType, pair<ea_t, sval_t> > &DeltaCopy2) const {
		STARSOpndType Opnd1 = DeltaCopy1.first;
		STARSOpndType Opnd2 = DeltaCopy2.first;
		if (IsEqOp(Opnd1, Opnd2)) {
			// If definitions are of the same operand, sort by instruction address of the definition.
			return (DeltaCopy1.second.first < DeltaCopy2.second.first);
		}
		if (Opnd1.type != Opnd2.type)
			return (Opnd1.type < Opnd2.type);
		switch (Opnd1.type) {
			case o_void: return false;
			case o_reg: return MDLessRegOpnd(Opnd1, Opnd2);
			case o_mem: return (Opnd1.addr < Opnd2.addr);
			case o_phrase: if (Opnd1.hasSIB && Opnd2.hasSIB) return ((Opnd1.sib < Opnd2.sib) || ((Opnd1.sib == Opnd2.sib) && (Opnd1.specflag4 < Opnd2.specflag4)));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return MDLessReg(Opnd1.phrase, Opnd2.phrase); // no SIB bytes
			case o_displ: if (Opnd1.hasSIB && Opnd2.hasSIB)
							  return ((Opnd1.sib < Opnd2.sib) 
								|| ((Opnd1.sib == Opnd2.sib) && ((Opnd1.addr < Opnd2.addr) || (Opnd1.specflag4 < Opnd2.specflag4))));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return ((Opnd1.addr < Opnd2.addr)
								|| ((Opnd1.addr == Opnd2.addr) 
								&& MDLessReg(Opnd1.reg, Opnd2.reg))); // no SIB bytes
			case o_imm: return (Opnd1.value < Opnd2.value);
			case o_far:  // fall through to o_near case
			case o_near: return (Opnd1.addr < Opnd2.addr);
			case o_trreg:  // fall through
			case o_dbreg:  // fall through
			case o_crreg:  // fall through
			case o_fpreg:  // fall through
			case o_mmxreg: // fall through
			case o_xmmreg: return (Opnd1.reg < Opnd2.reg); // no subword regs to deal with

			default: msg("ERROR: Unknown operand type in LessStackDeltaCopy.\n"); return false;
		}; // end switch (Opnd1.type)
	} // end operator
}; // end class LessStackDeltaCopy

// Get the size in bytes of the data type of an operand.
size_t GetOpDataSize(STARSOpndType DataOp);

// SMP will operate on a doubly linked list of instructions, which
//  will be grouped into basic blocks. We will include info about each
//  instruction that can not all be obtained easily in IDA Pro, and
//  is not all obtained easily from one place in IDA Pro. This 
//  information will include all that we need to perform data flow 
//  analyses. Because the LABEL and CASE could be applied to other
//  kinds of instructions during analysis, we define the enum values
//  as non-overlapping bits. This type might be stored as an int
//  in the future as a result of this design decision.
enum SMPitype {
	DEFAULT = 0,     // most instructions in middle of basic block
	LABEL = 1,       // instr also has a label (can be jump target) UNUSED
	CASE = 2,        // instr also has case label (is inside switch) UNUSED
	JUMP = 4,        // unconditional branch
	COND_BRANCH = 8, // conditional branch
	INDIR_JUMP = 16, // indirect unconditional branch,
	CALL = 32,       // direct call
	INDIR_CALL = 64, // indirect call
	RETURN = 128,    // function return OR TAIL CALL
	HALT = 256       // execution stops
};

// A primary goal of data flow analysis in SMP will be to identify
//  operands as numeric or pointer type. If an instruction definitely
//  produces a numeric result, an annotation could inform the mmStrata
//  run time monitor to simply record 'numeric' for the result register
//  and otherwise do nothing. This could greatly cut down the overhead
//  at run time. If the result register already has an mmStrata metadata
//  type of numeric, then the annotation could tell mmStrata to do nothing
//  at all, which is even better. In order to emit these annotations, we
//  will track all operands in a function as to their basic SMP types.
#define PROF_BASE 128
enum SMPOperandType { // What type is a given register or memory operand?
	UNINIT = 0,      // Operand type not yet analyzed; type lattice top
	NUMERIC = 1,     // Definitely holds non-pointer value (char, int, etc.)
	CODEPTR = 2,     // Definitely holds a code address; mmStrata considers this NUMERIC
	POINTER = 4,     // Definitely holds a data address
	STACKPTR = 8,    // Definitely holds a stack data address (refinement of POINTER)
	GLOBALPTR = 16,   // Definitely holds a global static data address (refinement of POINTER)
	HEAPPTR = 32,     // Definitely holds a heap data address (refinement of POINTER)
	PTROFFSET = 64,   // Offset from a pointer, e.g. a difference between two pointers
	NEGATEDPTR = 65,  // Temporary ptr arithmetic leaves NUMERIC-POINTER; should get +POINTER later to make PTROFFSET
	UNKNOWN = 96,     // Might hold an address, might not (Bad!); type lattice bottom
	PROF_NUMERIC = NUMERIC + PROF_BASE, // NUMERIC, determined using profiler input
	PROF_CODEPTR = CODEPTR + PROF_BASE, // CODEPTR, determined using profiler input
	PROF_POINTER = POINTER + PROF_BASE, // POINTER, determined using profiler input
	PROF_STACKPTR = STACKPTR + PROF_BASE, // STACKPTR, determined using profiler input
	PROF_GLOBALPTR = GLOBALPTR + PROF_BASE, // GLOBALPTR, determined using profiler input
	PROF_HEAPPTR = HEAPPTR + PROF_BASE, // HEAPPTR, determined using profiler input
	PROF_PTROFFSET = PTROFFSET + PROF_BASE, // PTROFFSET, determined using profiler input
	PROF_NEGATEDPTR = NEGATEDPTR + PROF_BASE, // NEGATEDPTR, determined using profiler input
	PROF_UNKNOWN = UNKNOWN + PROF_BASE // UNKNOWN, determined using profiler input
};

// Macros to query and compare SMPOperandType values.

// Is OpType any kind of pointer to data, whether profile derived or not?
#define IsDataPtr(OpType) (((OpType & (~PROF_BASE)) >= POINTER) && ((OpType & (~PROF_BASE)) <= HEAPPTR))

// Is OpType one of the refined data pointer types (stack, global, heap)?
#define IsRefinedDataPtr(OpType) (((OpType & (~PROF_BASE)) >= STACKPTR) && ((OpType & (~PROF_BASE)) <= HEAPPTR))

// Is OpType the lattice bottom (overdefined) type UNKNOWN, profile derived or not?
#define IsUnknown(OpType) ((OpType & (~PROF_BASE)) == UNKNOWN)

// Is OpType1 the same type as OpType2, ignoring profile derivation?
#define IsEqType(OpType1, OpType2)  ((OpType1 & (~PROF_BASE)) == (OpType2 & (~PROF_BASE)))
#define IsNotEqType(OpType1, OpType2)  ((OpType1 & (~PROF_BASE)) != (OpType2 & (~PROF_BASE)))

// Is OpType some kind of NUMERIC (i.e. NUMERIC or CODEPTR)?
#define IsNumeric(OpType) (IsEqType(OpType, NUMERIC) || IsEqType(OpType, CODEPTR))

// Is OpType a profile derived type?
#define IsProfDerived(OpType)  (0 != (OpType & PROF_BASE))

// Make OpType into its equivalent profile derived type.
#define MakeProfDerived(OpType) ((SMPOperandType)(OpType | PROF_BASE))

// Meet function over any two types in the type lattice.
SMPOperandType SMPTypeMeet(SMPOperandType Type1, SMPOperandType Type2);

// Meet function for SCCP constant propagation; updates NewConstStruct
void STARSConstantTypeMeet(struct STARS_SCCP_Const_Struct OldConstStruct, struct STARS_SCCP_Const_Struct &NewConstStruct);

// Enumeration to keep track of whether the metadata for each DEF is
//  used, unused, redundant, etc.
enum SMPMetadataType {
	DEF_METADATA_UNANALYZED = 0, // have not finished analyzing
	DEF_METADATA_UNUSED = 1,   // will not be used in a later store inst
	DEF_METADATA_USED = 2,     // will be used in bounds check for a store
	DEF_METADATA_USED_MEMWRITE = 4, // explanation of why DEF_METADATA_USED bit is set; reaches memory write
	DEF_METADATA_USED_CALL = 8, // explanation of why DEF_METADATA_USED bit is set; reaches call
	DEF_METADATA_USED_RETURN = 16, // explanation of why DEF_METADATA_USED bit is set; reaches return
	DEF_METADATA_REDUNDANT = 32, // DEF reg already has the same metadata
								// it would receive by updating from this inst
	DEF_METADATA_PROF_REDUNDANT = 64  // redundant, based on profiler-derived types
};

// Enumeration for constant propagation type lattice.
enum STARSConstantValueType {
	STARS_CONST_TOP = 0,  // top of lattice, uninitialized
	STARS_CONST_HAS_VALUE = 1, // Found 1 constant value so far; value is in SMPFunction::ConstantDefs
	STARS_CONST_BOTTOM = 2  // bottom of lattice, more than one value possible
};

struct STARS_SCCP_Const_Struct {
	enum STARSConstantValueType ConstType;
	uval_t ConstValue;   // holds valid value only if ConstType == STARS_CONST_HAS_VALUE.
};

// Encoding of info for fine-grained analysis of stack frames, integer error analysis, etc.
struct FineGrainedInfo {
	unsigned short SizeInfo;  // 8-bit, 16-bit, ...128-bit, int, float, string
	unsigned short SignMiscInfo; // signed, unsigned, read, written, FP-relative, etc
};

// Masks for FineGrainedInfo.SignMiscInfo
#define FG_MASK_SIGNED 1
#define FG_MASK_UNSIGNED 2
#define FG_MASK_INCONSISTENT_SIGN 3
#define FG_MASK_SIGNEDNESS_BITS 3  // better name for some uses
#define FG_MASK_READ 4
#define FG_MASK_WRITTEN 8
#define FG_MASK_SP_RELATIVE 16
#define FG_MASK_FP_RELATIVE 32
#define FG_MASK_ADDRESS_TAKEN 64
#define FG_MASK_INDEXED_ACCESS 128

// Masks for FineGrainedInfo.SizeInfo
#define FG_MASK_INTEGER 1
#define FG_MASK_STRING 2  // string data
#define FG_MASK_FLOAT_MMX 4  // floating point or MMX/SSE/packed
// remaining bit fields are the bit widths, e.g. 8 => 8-bit, ... 128 => 128-bit
#define FG_MASK_BITWIDTH_8 8
#define FG_MASK_BITWIDTH_16 16
#define FG_MASK_BITWIDTH_32 32
#define FG_MASK_BITWIDTH_64 64
#define FG_MASK_BITWIDTH_128 128
#define FG_MASK_BITWIDTH_256 256
#define FG_MASK_BITWIDTH_80 80   // two bits are set here
#define FG_MASK_BITWIDTH_96 512
#define FG_MASK_BITWIDTH_48 1024
#define FG_MASK_BITWIDTH_24 2048
#define FG_MASK_BITWIDTH_FIELDS (8|16|32|64|128|256|512|1024|2048)
// miscellaneous type info
#define FG_MASK_CODEPOINTER 4096
#define FG_MASK_DATAPOINTER 8192

// We don't worry about truncation unless we see operands less than normal width.
//  This would change if we port to 64-bit machines.
#define MD_NORMAL_MACHINE_BITWIDTH 32
#define MD_NORMAL_BITWIDTH_MASK FG_MASK_BITWIDTH_32

// Return one of the bit width masks above for the current operand.
unsigned short ComputeOperandBitWidthMask(STARSOpndType CurrOp, size_t DataSize);

// Compute largest bit width from a SignMiscInfo bit mask.
size_t LargestBitWidthFromMask(unsigned short WidthTypeInfo);

// Map the two sign bits above into four strings for annotation output.
extern const char *SignednessStrings[4];
// Version with NOFLAG prefixed for the load effective address opcodes
//  that can overflow silently without setting any flags.
extern const char *LeaSignednessStrings[4];

// We maintain a list of the caller-saved regs for the current binary's ABI.
//  This differs from 32-bit to 64-bit x86 binaries, as well as across other ISAs.
void MDInitializeCallerSavedRegs(void);
list<uint16>::iterator GetFirstCallerSavedReg(void);
list<uint16>::iterator GetLastCallerSavedReg(void);

// We maintain a list of the argument-passing regs for the current binary's ABI.
//  This differs from 32-bit to 64-bit x86 binaries, as well as across other ISAs.
//  The list is in order of argument position number. For x86-64, this means EDI,
//  ESI, EDX, ECX, R8, R9.
void MDInitializeArgumentRegs(void);
list<uint16>::iterator GetFirstArgumentReg(void);
list<uint16>::iterator GetLastArgumentReg(void);

class DisAsmString {
public:
	DisAsmString(void);
	char *GetDisAsm(ea_t InstAddr, bool MarkerInst = false);
	void SetMarkerInstText(ea_t InstAddr);

private:
	ea_t CurrAddr; // cached string corresponds to this instruction address
	ssize_t StringLen; // string length returned by IDA Pro
	char CachedDisAsm[MAXSMPDISASMSTR];
};

// Masks to set bits to use as booleans1 in DefOrUse.
#define DEFORUSE_SET_INDIRECT_WRITE 0x01
#define DEFORUSE_SET_NO_TRUNCATION 0x02
#define DEFORUSE_SET_NO_OVERFLOW 0x04
#define DEFORUSE_SET_UNUSED2 0x08
#define DEFORUSE_SET_UNUSED3 0x10
#define DEFORUSE_SET_UNUSED4 0x20
#define DEFORUSE_SET_UNUSED5 0x40
#define DEFORUSE_SET_UNUSED6 0x80
#define DEFORUSE_RESET_INDIRECT_WRITE 0xfe
#define DEFORUSE_RESET_NO_TRUNCATION 0xfd
#define DEFORUSE_RESET_NO_OVERFLOW 0xfb
#define DEFORUSE_RESET_UNUSED2 0xf7
#define DEFORUSE_RESET_UNUSED3 0xef
#define DEFORUSE_RESET_UNUSED4 0xdf
#define DEFORUSE_RESET_UNUSED5 0xbf
#define DEFORUSE_RESET_UNUSED6 0x7f

class DefOrUse {
public:
	// Constructors
	DefOrUse(void);
	DefOrUse(STARSOpndType Ref, SMPOperandType Type = UNINIT, int SSASub = SMP_SSA_UNINIT);
	DefOrUse(const DefOrUse &CopyIn);

	// Operators
	DefOrUse &operator=(const DefOrUse &rhs);

	// Get methods
	inline STARSOpndType GetOp(void) const { return Operand; };
	inline SMPOperandType GetType(void) const { return OpType; };
	inline int GetSSANum(void) const { return SSANumber; };
	inline SMPMetadataType GetMetadataStatus(void) const { return MetadataStatus; };

	// Query methods.
	inline bool HasIndirectWrite(void) const { return (booleans1 & DEFORUSE_SET_INDIRECT_WRITE); };
	inline bool DoesNotTruncate(void) const { return (booleans1 & DEFORUSE_SET_NO_TRUNCATION); };
	inline bool DoesNotOverflow(void) const { return (booleans1 & DEFORUSE_SET_NO_OVERFLOW); };

	// Set methods
	inline void SetOp(STARSOpndType NewOp) { Operand = NewOp; };
	inline void SetSSANum(int Num) { SSANumber = Num; };
	void SetType(SMPOperandType Type, const SMPInstr* instr); 
	void SetMetadataStatus(SMPMetadataType NewStatus);
	inline void SetIndWrite(bool IndMemWrite) { booleans1 |= DEFORUSE_SET_INDIRECT_WRITE; };
	inline void SetNoTruncation(bool NoTruncFlag)  { booleans1 |= DEFORUSE_SET_NO_TRUNCATION; };
	inline void SetNoOverflow(bool NoOverflowFlag)  { booleans1 |= DEFORUSE_SET_NO_OVERFLOW; };

	// Printing methods
	void Dump(void) const;
private:
	// Data
	STARSOpndType Operand;
	int SSANumber;
	SMPOperandType OpType;
	SMPOperandType NonSpeculativeOpType;
	SMPMetadataType MetadataStatus;
	unsigned char booleans1; // holds bit-flags
#if 0  // Moved flags into booleans1.
	bool IndWrite; // Does an indirect memory write occur within the chain
	               //  starting at this DEF? (always false for a USE)
	bool NoTruncation; // USE appears to be a truncation of its original DEF, but is not.
	bool NoOverflow; // DEF appears to be an overflowing computation, but is not.
#endif
}; // end of class DefOrUse

// Comparison operator class to permit use of DefOrUse type in sets.
class LessDefUse {
public:
	bool operator()(const DefOrUse &Ref1, const DefOrUse &Ref2) const {
		set<STARSOpndType, LessOp>::value_compare OpComp = this->DummySet.value_comp(); // LessOp()
		return OpComp(Ref1.GetOp(), Ref2.GetOp());
	}
private:
	set<STARSOpndType, LessOp> DummySet;
};

// Same class is used for both a DEF set and a USE set.
class DefOrUseSet {
public:
	// Constructors
	DefOrUseSet(void);
	// Destructor.
	~DefOrUseSet();

	// Get methods
	inline size_t GetSize(void) const { return (size_t) Refs.size(); };
	inline set<DefOrUse, LessDefUse>::iterator GetFirstRef(void) { return Refs.begin(); };
	inline set<DefOrUse, LessDefUse>::iterator GetLastRef(void) { return Refs.end(); };
	set<DefOrUse, LessDefUse>::iterator FindRef(STARSOpndType SearchOp);
	
	// Set methods
	// Insert a new DEF or USE; must be new, insert must succeed else we assert.
	set<DefOrUse, LessDefUse>::iterator InsertRef(DefOrUse Ref); 
	void SetRef(STARSOpndType Ref, SMPOperandType Type = UNINIT, int SSASub = SMP_SSA_UNINIT);
	set<DefOrUse, LessDefUse>::iterator SetOp(set<DefOrUse, LessDefUse>::iterator CurrRef, STARSOpndType NewOp);
	set<DefOrUse, LessDefUse>::iterator SetSSANum(STARSOpndType CurrOp, int NewSSASub);
	set<DefOrUse, LessDefUse>::iterator SetType(STARSOpndType CurrOp, SMPOperandType Type, const SMPInstr* Instr);
	set<DefOrUse, LessDefUse>::iterator SetMetadata(STARSOpndType CurrOp, SMPMetadataType Status);
	// Change the indirect write status for a reference:
	set<DefOrUse, LessDefUse>::iterator SetIndWrite(STARSOpndType CurrOp, bool IndWriteFlag);
	// Change the ignore apparent truncation flag for a reference:
	set<DefOrUse, LessDefUse>::iterator SetNoTruncation(STARSOpndType CurrOp, bool NoTruncFlag);
	// Change the ignore apparent overflow flag for a reference:
	set<DefOrUse, LessDefUse>::iterator SetNoOverflow(STARSOpndType CurrOp, bool NoOverflowFlag);
	inline void clear(void) { Refs.clear(); return; };
	inline void EraseRef(set<DefOrUse, LessDefUse>::iterator RefIter) {
		Refs.erase(RefIter); return;
	}

	// Printing methods
	void Dump(void) const;

	// Analysis methods
	bool TypesAgreeNoFlags(void); // Are all types consistent, ignoring flag registers?
private:
	// Data
	set<DefOrUse, LessDefUse> Refs;  // Defined or used operand with type and SSA subscript
};  // end class DefOrUseSet

// Same class is used for both a DEF list and a USE list.
class DefOrUseList {
public:
	// Constructors
	DefOrUseList(void);

	// Get methods
	DefOrUse GetRef(size_t index) const;
	inline size_t GetSize(void) const { return (size_t) Refs.size(); };
	inline DefOrUse *GetRefNum(size_t index) { return &Refs[index]; };
	inline vector<DefOrUse>::iterator GetFirstRef(void) { return Refs.begin(); };
	inline vector<DefOrUse>::iterator GetLastRef(void) { return Refs.end(); };
	inline int GetRefSSANum(size_t index) const { return Refs.at(index).GetSSANum(); };
	inline SMPOperandType GetRefType(size_t index) const { return Refs.at(index).GetType(); };

	// Set methods
	void SetRef(STARSOpndType Ref, SMPOperandType Type = UNINIT, int SSASub = SMP_SSA_UNINIT);
	void SetSSANum(size_t index, int NewSSASub);
	void SetType(size_t index, SMPOperandType Type, const SMPInstr* Instr);
	inline void clear(void) { Refs.clear(); return; };
	// Printing methods
	void Dump(void) const;
	// Analysis methods
	void EraseDuplicates(void);  // in case SMPInstr::MDFixupDefUseLists() adds duplicate
private:
	// Data
	vector<DefOrUse> Refs;  // Defined or used operand with type, SSA subscript, etc.
};  // end class DefOrUseList

class SMPPhiFunction {
public:
	// Constructors
	SMPPhiFunction(int GlobIndex, const DefOrUse &Def);

	// Get methods
	inline int GetIndex(void) const { return index; };
	inline size_t GetPhiListSize(void) const { return this->SubscriptedOps.GetSize(); };
	DefOrUse GetDefCopy(void) const;
	inline DefOrUse GetPhiRef(size_t i) const { return this->SubscriptedOps.GetRef(i); };
	inline DefOrUse *GetRefNum(size_t index) { return SubscriptedOps.GetRefNum(index); };
	inline vector<DefOrUse>::iterator GetFirstOp(void) { return SubscriptedOps.GetFirstRef(); };
	inline vector<DefOrUse>::iterator GetLastOp(void) { return SubscriptedOps.GetLastRef(); };
	inline STARSOpndType GetAnyOp(void) const { return DefName.GetOp(); };
	inline int GetDefSSANum(void) const { return DefName.GetSSANum(); };
	inline int GetUseSSANum(size_t index) const { return SubscriptedOps.GetRefSSANum(index); };
	inline SMPOperandType GetDefType(void) const { return DefName.GetType(); };
	inline SMPOperandType GetUseType(size_t index) const { return SubscriptedOps.GetRefType(index); };
	inline SMPMetadataType GetDefMetadata(void) const { return DefName.GetMetadataStatus(); };

	// Set methods
	void PushBack(DefOrUse Ref); // add inputs to phi function
	void SetSSADef(int NewSSASub);
	void SetSSARef(size_t index, int NewSSASub);
	void SetDefType(SMPOperandType Type, const SMPInstr* instr);
	void SetRefType(size_t index, SMPOperandType Type, const SMPInstr* instr);
	void SetDefMetadata(SMPMetadataType Status);

	// Query methods
	bool HasTypedUses(void); // true ==> at least one USE is not UNINIT type
	// Printing methods
	void Dump(void) const;
	// Analysis methods
	SMPOperandType ConditionalMeetType(SMPBasicBlock *CurrBlock) const; // meet operator over USE types
		// WARNING: This method propagates types, which can cause the Phi iterator in the enclosing SMPBasicBlock to change
		//  and no longer be equal to the "this" pointer in this SMPPhiFunction. Need to re-do FindPhi() to get non-stale iterator.

private:
	int index;  // index into SMPFunction::BlocksDefinedIn (i.e. global name index)
	DefOrUse DefName; // output of phi function
	DefOrUseList SubscriptedOps; // ordered list of inputs to phi function
}; // end class SMPPhiFunction

// Comparison function to permit sorting phi functions and keeping them in sets.
class LessPhi {
public:
	bool operator() (const SMPPhiFunction &Phi1, const SMPPhiFunction &Phi2) const {
		return (Phi1.GetIndex() < Phi2.GetIndex());
	}
}; // end of class LessPhi

// In SSA computations, we are storing the GlobalNames index into the op_t fields
//  n, offb, and offo. This function extracts an unsigned int from these three 8-bit
//  fields.
unsigned int ExtractGlobalIndex(STARSOpndType GlobalOp);
void SetGlobalIndex(STARSOpndType *TempOp, size_t index);

class SMPDefUseChain {
public:
	// Constructors
	SMPDefUseChain(void);
	SMPDefUseChain(STARSOpndType Name, ea_t Def = BADADDR);
	// Get methods
	inline STARSOpndType GetName(void) const { return SSAName; };
	inline ea_t GetDef(void) const { return (ea_t) RefInstrs.at(0); };
	inline ea_t GetUse(size_t index) const { return (ea_t) RefInstrs.at(index + 1); };
	inline ea_t GetLastUse(void) const {
		if (RefInstrs.size() > 1) 
			return (ea_t) RefInstrs.back();
		else
			return (ea_t) BADADDR;
	};
	inline size_t GetNumUses(void) const { return RefInstrs.size() - 1; };

	// Query methods.
	inline bool HasIndirectWrite(void) const { return IndWrite; };

	// Set methods
	void SetName(STARSOpndType Name);
	void SetDef(ea_t Def);
	void PushUse(ea_t Use);
	void SetIndWrite(bool IndMemWrite);
	// Printing methods
	void Dump(int SSANum = (-1)) const;
private:
	STARSOpndType SSAName;  // What variable is defined and used in the chain?
	bool IndWrite; // Does an indirect memory write occur within the chain?

	// Keep DEF and USE addrs as unsigned short offsets from BaseAddr of block.
	//  The BaseAddr of the block is 1 less than the first inst addr of the block,
	//  so that this value can be used as a flag to indicate that the value was
	//  live-in and upward-exposed upon block entry and does not have a real DEF
	//  in the block for that SSANum.
	vector<unsigned short> RefInstrs; // First is always DEF, rest are USE.
}; // end class SMPDefUseChain

class SMPDUChainArray {
public:
	// Constructor
	SMPDUChainArray(void);
	SMPDUChainArray(STARSOpndType Name, ea_t FirstAddrMinusOne);
	// Get methods.
	inline STARSOpndType GetName(void) const { return SSAName; };
	inline ea_t GetDef(int SSANum) const { return BaseAddr + DUChains.at(SSANum).GetDef(); };
	inline ea_t GetUse(int SSANum, size_t index) const { return BaseAddr + DUChains.at(SSANum).GetUse(index); };
	ea_t GetLastUse(int SSANum) const;
	inline size_t GetNumUses(int SSANum) const { return DUChains.at(SSANum).GetNumUses(); };
	inline size_t GetSize(void) const { return DUChains.size(); };
	// Set methods.
	void SetName(STARSOpndType Name, ea_t FirstAddrMinusOne);
	inline void PushUse(int SSANum, ea_t UseAddr) { DUChains.at(SSANum).PushUse(UseAddr - BaseAddr); };
	inline void PushChain(SMPDefUseChain NewChain) { DUChains.push_back(NewChain); };
	inline void SetIndWrite(int SSANum, bool NewFlag) { DUChains.at(SSANum).SetIndWrite(NewFlag); };
	// Query methods
	inline bool HasIndirectWrite(int SSANum) const { return DUChains.at(SSANum).HasIndirectWrite(); };
	// Printing methods.
	void Dump(void) const;
private:
	// Data
	vector<SMPDefUseChain> DUChains; // indexed by SSA number for local chains, by chain # for global chains
	STARSOpndType SSAName; // What variable is used in all chains in the array?
	ea_t BaseAddr; // one less than first addr in basic block
}; // end class SMPDUChainArray

class SMPCompleteDUChains {
public:
	// Printing methods.
	void Dump(void) const;
	// Data (public for convenience)
	vector<SMPDUChainArray> ChainsByName; // indexed by name index
}; // end class SMPCompleteDUChains

// Our own bitset class. The C++ bitset class cannot be dynamically sized. The vector<bool> specialization
//  of the C++ STL vector container does not behave like other containers and is in danger of being deprecated.
class STARSBitSet {
public:
	// Constructors
	STARSBitSet();
	// Get methods
	bool GetBit(size_t BitIndex) const;
	// Set methods
	void AllocateBits(size_t Size); // allocate Size bits, initialized to zero.
	void SetBit(size_t BitIndex);
	void ResetBit(size_t BitIndex);
	// Query methods
	bool IsAnyBitSet(void) const; // Returns false if all bits are zero, true otherwise.

private:
	// Data members.
	size_t BitLimit; // number of bits
	vector<unsigned char> STARSBits;
	// Methods
}; // end class STARSBitSet

// Initialization routine for DFA category.
extern SMPitype DFACategory[];
void InitDFACategory(void);

// Initialization routine for Type category.
extern int SMPTypeCategory[];
void InitTypeCategory(void);

// Initializations for CPU flags DEF/USE.
extern bool SMPDefsFlags[];
extern bool SMPUsesFlags[];
void InitSMPDefsFlags(void);
void InitSMPUsesFlags(void);

// Initialize the FG info for the return register from any library function
//  whose name implies that we know certain return values (e.g. atoi() returns
//  a signed integer, while strtoul() returns an unsigned long).
void GetLibFuncFGInfo(string FuncName, struct FineGrainedInfo &InitFGInfo);

// Initialize the lookup maps that are used to define the FG info that can
//  be inferred from a library function name.
void InitLibFuncFGInfoMaps(void);

// Init map of system or library call name to the annotation substring that
//  guides saturating arithmetic or other continuation policies in 
//  the case of integer error detection of a value passed to that call.
void InitIntegerErrorCallSinkMap(void);

// Return sink string for call name from the sink map.
// If we don't care find the call name, we return an empty string.
void GetSinkStringForCallName(string CalleeName, string &SinkString);

// Init map of system or library call name to the argument number that
//  should have an unsigned value and should be guarded from the
//  signedness error that results from copying a signed value
//  into the outgoing argument. Argument numbers are zero-based; see bit defs below.
void InitUnsignedArgPositionMap(void);

// Define argument position bits to put into various Arg Position Maps.
// A system or library call name could have more than one argument
//  position to protect from signedness errors.
#define STARS_ARG_POS_0 0x1
#define STARS_ARG_POS_1 0x2
#define STARS_ARG_POS_2 0x4
#define STARS_ARG_POS_3 0x8
#define STARS_ARG_POS_4 0x16

// Return unsigned argument position bitmap for call name from the unsigned arg position map.
//  We will return 0 when there is no argument to worry about
//  for a particular library or system call name.
void GetUnsignedArgPositionsForCallName(string CalleeName, unsigned int &ArgPosBits);

// Init map of system or library call name to the argument number that is dangerous
//  to accept from a user-tainted source. Lots of overlap with UnsignedArgPositionMap.
void InitTaintWarningArgPositionMap(void);

// Return unsigned argument position bitmap for call name from the taint warning arg position map.
//  We will return 0 when there is no argument to worry about
//  for a particular library or system call name.
void GetTaintWarningArgPositionsForCallName(string CalleeName, unsigned int &ArgPosBits);

// Init map of system or library call name to the argument number that
//  should have a POINTER value.
void InitPointerArgPositionMap(void);

// Return POINTER argument position bitmap for call name from the taint warning arg position map.
//  We will return 0 when there is no POINTER argument to worry about
//  for a particular library or system call name.
void GetPointerArgPositionsForCallName(string CalleeName, unsigned int &ArgPosBits);

// Utility to count bits set in an unsigned int, e.g. ArgPosBits.
unsigned int CountBitsSet(unsigned int ArgPosBits);

#endif
